/*-------------------------------------------------------------------*/
/* CBSPLIT.C      Timo Kaukoranta                                    */
/*                                                                   */
/*                                                                   */
/* - Splitting Neighbour method                                      */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBSPLIT"
#define VersionNumber  "V.0.37"
#define LastUpdated    "13.11.00"

/* ----------------------------------------------------------------- */

#include <string.h>
#ifdef UNIX
#include <sys/time.h>
#endif

#include <time.h>
#include <values.h>

#define  FACTFILE  "cbsplit.fac"

#include "parametr.c"

#include "file.h"
#include "memctrl.h"
#include "random.h"
#include "sa.h"
#include "sortcb.h"

#include "split.h"


/*--------------------  Basic type definitions -----------------------*/

/* None. */

/*-----------------------------  M i s c  ----------------------------*/

static int ResultCBSize       = 0;

#define MAXFILENAME   64


/* =================  T I M E   F U N C T I O N S  ================= */


/* Use 'gettimeofday' for more accurate measure of time. */
#if 1

/* ----------------------------------------------------------------- */

double SetWatch(double* start)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);

  return( *start = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) );
}

/* ----------------------------------------------------------------- */

void PrintTime(double start)
{
  struct timeval tv;
  double elapsed;

  gettimeofday(&tv, NULL);
  elapsed = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) - start;
  switch( Value(ShowProgress) )
    {
    case 0:  break;
    case 1:  printf("%6.1f ", elapsed);
             break;
    default: printf("Time: %6.1f\n", elapsed);
             break;
    }
}

/* ----------------------------------------------------------------- */

#else

long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


void PrintTime(long start)
{
  time_t tmp;
  long   elapsed = (long)time(&tmp) - start;

  switch( Value(ShowProgress) )
    {
    case 0:  break;
    case 1:  printf("%6li ", elapsed);
             break;
    default: printf("Time:%6li\n", elapsed);
             break;
    }
}

#endif

/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("This program generates a codebook by SPLIT method.\n");
  printf("Usage: %s [%coption] <training-set> [initial-codebook] <result-codebook>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s set_of_4 gla glasplit\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char  TSetName[],
                               char  InitialCBName[],
                               char  ResultCBName[])
{
  if( Value(ShowProgress) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Training Set:     %s\n", TSetName);
    printf("Initial Codebook: %s\n", (InitialCBName[0] == 0x00
                                     ? "Centroid of the TRAINING SET"
                                     : InitialCBName) );
    printf("Result Codebook:  %s\n", ResultCBName);
    printf("Result CB size:   %u\n", ResultCBSize);
    PrintSelectedOptions();
    printf("\n");
    }
}


/*=========================  CODEBOOK HANDLING  ============================*/


static void GenerateInitialCodebook(char*         InitialCBName,
                                    TRAININGSET*  TS,
                                    CODEBOOK*     CB,
                                    PARTITIONING* P)
{
  if( InitialCBName[0] != 0x00 ) /* Initial CB is given. */
    {
    ReadCodebook(InitialCBName, CB);
    if( Value(CodebookSizeChange) > 0 )
      {
      ResultCBSize = BookSize(CB) + Value(CodebookSizeChange);
      }
    else
      {
      ResultCBSize = Value(CodebookSize);
      }

    IncreaseCodebookSize(CB, ResultCBSize);
    CreateNewPartitioning(P, TS, ResultCBSize);
    GenerateOptimalPartitioning(TS, CB, P);

    AddGenerationMethod(CB, "SPLIT");
    }
  else /* No initial CB */
    {
    /* Use the centroid of the TS as a initial CB. */
    if( Value(CodebookSizeChange) > 0 )
      {
      ResultCBSize = 1 + Value(CodebookSizeChange);
      }
    else
      {
      ResultCBSize = Value(CodebookSize);
      }

    CreateNewCodebook(CB, 1, TS);
    CreateNewPartitioning(P, TS, 1);

    /* Just centroid of TS. */
    PartitionCentroid(P, 0, &Node(CB, 0));

    AddGenerationMethod(CB, "SPLIT");
    }
}


/* ----------------------------------------------------------------- */


static void SaveResultCodebook(char*     ResultCBName,
                               CODEBOOK* CB)
{
  int ndups;

  if( (ndups = DuplicatesInCodebook(CB)) > 0 )
    {
    printf("WARNING: Final CB contains %i duplicates.\n", ndups);
    }
  SortCodebook(CB, FREQ_DESCENDING);
  WriteCodebook(ResultCBName, CB, Value(OverWrite));
}


/* ================================================================= */


static void Initialize(TRAININGSET*  TS,
                       CODEBOOK*     CB,
                       SASchedule*   SAS)
{
  int ndups;

  if( (ndups = DuplicatesInCodebook(CB)) > 0 )
    {
    if( Value(ShowProgress) >= 1 )
      {
      printf("ERROR: Init. CB does contain %i duplicates.\n", ndups);
      exit( -1 );
      }
    }

  SetSplitParameters(Value(ShowProgress),
                     Value(PartitionErrorType),
                     Value(NewVectors),
                     Value(NewVectorsHeuristically),
                     Value(Hyperplane),
                     Value(HyperplanePivot),
                     Value(HyperplaneLine),
                     Value(PartitionRemapping),
                     Value(LocalGLA),
                     Value(LocalGLAIterations),
                     Value(GLAIterations));
  InitializeSASchedule(Value(TemperatureDistribution),
                       Value(TemperatureFunction),
                       Value(TemperatureChange),
                       (float)Value(TemperatureInitial) * (float)(CB->MaxValue) / 100.0,
                       (Value(VectorRandomizing) == RandomCentroid ||
                        Value(VectorRandomizing) == RandomBoth),
                       (Value(VectorRandomizing) == RandomTrainingSet ||
                        Value(VectorRandomizing) == RandomBoth),
                       SAS);
}


/* ================================================================= */


static void ShutUp(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* P)
{
  FreeCodebook(TS);
  FreeCodebook(CB);
  FreePartitioning(P);
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  CODEBOOK      CB;
  TRAININGSET   TS;
  PARTITIONING  P;
  char          TSetName[MAXFILENAME] = { 0x00 };
  char          InitialCBName[MAXFILENAME] = { 0x00 };
  char          ResultCBName[MAXFILENAME] = { 0x00 };
  SASchedule    SAS;
  double        FinalError;
  double        watch;

  ParameterInfo paraminfo[3] = { { TSetName,      FormatNameTS, 0, INFILE  },
                                 { InitialCBName, FormatNameCB, 1, INFILE  },
                                 { ResultCBName,  FormatNameCB, 0, OUTFILE } };

  ParseParameters(argc, argv, 3, paraminfo);

  initrandom(Value(RandomSeed));

  ReadTrainingSet(TSetName, &TS);

  SetWatch(&watch);

  GenerateInitialCodebook(InitialCBName, &TS, &CB, &P);

  Initialize(&TS, &CB, &SAS);
  PrintOperatingInfo(TSetName, InitialCBName, ResultCBName);

  Split(&TS, &CB, &SAS, &P, ResultCBSize);

  PrintTime(watch);

  FinalError = AverageErrorCBFast(&TS, &CB, &P, MSE);
  if( Value(ShowProgress) >= 2 )
    {
    printf("Final dist. = %9.4f\n", PrintableError(FinalError, &CB));
    }
  else
    {
    printf("%9.4f ", PrintableError(FinalError, &CB));
    }

  SaveResultCodebook(ResultCBName, &CB);
  ShutUp(&TS, &CB, &P);
  checkmemory();

  return( 0 );
}

